mdmg_header = [
    "tax_id",
    "sample",
    "N_reads",
    "lambda_LR",
    "D_max",
    "mean_L",
    "mean_GC",
    "q",
    "A",
    "c",
    "phi",
    "rho_Ac",
    "valid",
    "asymmetry",
    "std_L",
    "std_GC",
    "D_max_std",
    "q_std",
    "phi_std",
    "A_std",
    "c_std",
    "lambda_LR_P",
    "lambda_LR_z",
    "forward_lambda_LR",
    "forward_D_max",
    "forward_D_max_std",
    "forward_q",
    "forward_q_std",
    "forward_phi",
    "forward_phi_std",
    "forward_A",
    "forward_A_std",
    "forward_c",
    "forward_c_std",
    "forward_rho_Ac",
    "forward_lambda_LR_P",
    "forward_lambda_LR_z",
    "forward_valid",
    "reverse_lambda_LR",
    "reverse_D_max",
    "reverse_D_max_std",
    "reverse_q",
    "reverse_q_std",
    "reverse_phi",
    "reverse_phi_std",
    "reverse_A",
    "reverse_A_std",
    "reverse_c",
    "reverse_c_std",
    "reverse_rho_Ac",
    "reverse_lambda_LR_P",
    "reverse_lambda_LR_z",
    "reverse_valid",
    "LR_All",
    "LR_ForRev",
    "LR_ForRev_All",
    "chi2_all",
    "chi2_forward",
    "chi2_reverse",
    "chi2_ForRev",
    "N_x=1_forward",
    "N_x=1_reverse",
    "N_sum_total",
    "N_sum_forward",
    "N_sum_reverse",
    "N_min",
    "k_sum_total",
    "k_sum_forward",
    "k_sum_reverse",
    "Bayesian_z",
    "Bayesian_D_max",
    "Bayesian_D_max_std",
    "Bayesian_A",
    "Bayesian_A_std",
    "Bayesian_q",
    "Bayesian_q_std",
    "Bayesian_c",
    "Bayesian_c_std",
    "Bayesian_phi",
    "Bayesian_phi_std",
    "Bayesian_rho_Ac",
    "Bayesian_significance",
]

valid_ranks = {
    "domain": "d__",
    "kingdom": "k__",
    "lineage": "l__",
    "phylum": "p__",
    "class": "c__",
    "order": "o__",
    "family": "f__",
    "genus": "g__",
    "species": "s__",
}